import pytest

from PyThermite import Index, Indexable, QueryExpr as Q, FilteredIndex

class TestClass(Indexable):
    def some_method(self):
        return "Hello from TestClass"


@pytest.fixture
def indexable():
    # Basic creation of an Indexable object.
    return Indexable()


@pytest.fixture
def index():
    return Index()


def test_non_nested(index: Index):
    for i in range(10):
        index.add_object(TestClass(id=1, num = i//2))

    grouped = index.group_by("num")

    for i in range(4):
        assert len(grouped[i].collect()) == 2

def test_non_nested_many_children(index: Index):
    vals = [0,1,2,3,4,5,6,7,8,9]
    for i in range(10):
        print(vals[i:])
        index.add_object(TestClass(id=1, num = i//2, nested = vals[i:]))

    grouped = index.group_by("nested")

    for i in vals:
        assert len(grouped[i].collect()) == i + 1

def test_nested_group_by(index: Index):
    for i in range(10):
        nested = TestClass(nest=True, num = i//2)
        index.add_object(TestClass(id=1, nested=nested))

    grouped = index.group_by("nested.num")

    zeros = grouped[0].collect()
    assert len(zeros) == 2
    assert zeros[0].nested.num == 0
    assert zeros[1].nested.num == 0

    grouped = index.group_by("nested.nest")

    all = grouped[True].collect()
    assert len(all) == 10

def test_nested_group_by_with_many_children(index: Index):
    for iteration in range(10):
        t = [i for i in range(iteration)]
        nested = TestClass(nest=True, children=t)
        index.add_object(TestClass(id=1, nested=nested))

    grouped = index.group_by("nested.children")

    for i in grouped:
        assert len(grouped[i].collect()) == 9-i

def test_group_by_one_to_many(index: Index):
    nested = TestClass(nest=True)
    for i in range(10):
        index.add_object(TestClass(id=i, nested=nested))

    grouped = index.group_by("nested.nest")

    all = grouped[True].collect()
    assert len(all) == 10

def test_group_by_many_to_many(index: Index):
    nested = [TestClass(num = i, nest=True, id = i) for i in range(10)]
    for i in range(10):
        index.add_object(TestClass(id=i, nested=nested))


    grouped = index.group_by("nested.num")
    # change any one from nested
    grouped[0].collect()[0].nested[0].idx = 100

    # all changes should reflect
    for ind in index.collect():
        assert ind.nested[0].idx == 100

def test_group_by_one_to_many_deregister(index: Index):
    nested = TestClass(nest=True)
    for i in range(10):
        index.add_object(TestClass(id=i, nested=nested))

    grouped = index.group_by("nested.nest")

    all = grouped[True].collect()
    assert len(all) == 10

    all[0].nested = "test"

    grouped = index.group_by("nested.nest")

    all = grouped[True].collect()
    assert len(all) == 9

def test_remove_and_reassign_tracked_list(index: Index):
    nested = [TestClass(num=i, id=i) for i in range(3)]
    
    objs = [TestClass(id=i, nested=nested) for i in range(5)]
    for obj in objs:
        index.add_object(obj)

    grouped = index.group_by("nested.num")
    assert len(grouped[0].collect()) == 5

    for obj in objs:
        obj.nested = []

    grouped_after_remove = index.group_by("nested.num")
    assert all(len(v.collect()) == 0 for _, v in grouped_after_remove.items())

    new_nested = [TestClass(num=100 + i, id=10 + i) for i in range(3)]
    for obj in objs:
        obj.nested = new_nested

    grouped_after_reassign = index.group_by("nested.num")
    first_group = grouped_after_reassign[100].collect()
    assert all(o.nested[0].num == 100 for o in first_group)

    new_nested[0].num = 999
    grouped_after_mutation = index.group_by("nested.num")
    first_group = grouped_after_mutation[999].collect()
    assert all(o.nested[0].num == 999 for o in first_group)




class Index:
    '''
    Index is a collection of Indexable objects that can be queried and filtered.
    It supports adding and removing objects, as well as querying via attribute filters or query expressions.
    Experemential Thread safe, but not officially supported yet.
    '''

    def collect() -> list[Indexable]: 
        '''
        collects all valid objects in the index and returns them as a list
        '''
    ...
    def add_object(obj: Indexable): 
        '''
        adds a single object to the index
        '''
    ...
    def add_object_many(objects: list[Indexable]):
        '''
        adds a multiple objects to the index
        '''
    ...
    def reduce(**kwargs): 
        '''
        removes items in place that do not match the given attribute filters
        Note that this modifies the current index in place and is less performant 
        than using a method that returns a filteredIndex such as reduced_query or reduced        
        '''
    ...
    def reduced(**kwargs) -> FilteredIndex:
        '''
        returns a FilteredIndex containing only items that match the given attribute filters
        '''
    ...
    def reduced_query(query: PyQueryExpr) -> FilteredIndex:
        '''
        returns a FilteredIndex containing only items that match the given query expression
        '''
    ...
    def get_by_attribute(**kwargs):
        '''
        a shorthand for reduced(**kwargs).collect()
        more performant than using reduced when you only need the collected results
        and not the FilteredIndex to further query
        '''    
    ...
    def union_with(other: Index):
        '''
        returns a new Index that is the union of this index and another index
        does not mutate the other index
        '''    
    ...

class FilteredIndex:
    '''
    FilteredIndex is a view into an Index with an allow list of items.
    It supports the same querying and filtering operations as Index
    but does not support adding or removing objects.
    '''
    def reduced(**kwargs) -> FilteredIndex:
        '''
        returns a FilteredIndex containing only items that match the given attribute filters
        '''    
    ...
    def reduced_query(query: PyQueryExpr) -> FilteredIndex:
        '''
        returns a FilteredIndex containing only items that match the given query expression
        '''
    ...
    def collect() -> list[Indexable]:
        '''
        collects all valid objects in the FilteredIndex and returns them as a list
        '''
    ...
    def rebase() -> Index:
        '''
        returns a new Index containing only the items in this FilteredIndex
        Only use this when a full Index is needed, as it is much less performant
        '''    
    ...

class PyQueryExpr:
    def and_(queries: list[PyQueryExpr]) -> PyQueryExpr:
        '''
        all subqueries must be true
        '''    
    ...
    def or_(queries: list[PyQueryExpr]) -> PyQueryExpr:
        '''
        at least one subquery must be true
        '''    
    ...
    def not_(queries: PyQueryExpr) -> PyQueryExpr:
        '''
        negates the subquery
        '''    
    ...

    def eq(attr: str, value: any) -> PyQueryExpr:
        '''
        attribute equals value
        '''    
    ...
    def in_(attr: str, values: list[any]) -> PyQueryExpr:
        '''
        attribute is in the list of values
        equivilant to multiple or'd eq queries
        '''
    ...
    def ne(attr: str, value: any) -> PyQueryExpr:
        '''
        attribute does not equal value
        '''    
    ...

    def gt(attr: str, value: int | float) -> PyQueryExpr:
        '''
        attribute is greater than value
        '''
    ...
    def ge(attr: str, value: int | float) -> PyQueryExpr:
        '''
        attribute is greater than or equal to value
        '''    
    ...
    def lt(attr: str, value: int | float) -> PyQueryExpr:
        '''
        attribute is less than value
        '''
    ...
    def le(attr: str, value: int | float) -> PyQueryExpr:
        '''
        attribute is less than or equal to value
        '''    
    ...
    def bt(attr: str, lower: int | float, upper: int | float) -> PyQueryExpr:
        '''
        attribute is between lower and upper, inclusive
        '''    
    ...

class Indexable:
    '''
    Base class for objects that can be indexed.
    All attributes will be indexed unless the attribute name is prefixed with an underscore
    Nested Indexable objects are supported and fully queryable via dot notation.
    '''
...





import pytest

from PyThermite import Index, Indexable, QueryExpr as Q, FilteredIndex

class TestClass(Indexable):
    def some_method(self):
        return "Hello from TestClass"


@pytest.fixture
def indexable():
    # Basic creation of an Indexable object.
    return Indexable()


@pytest.fixture
def index():
    return Index()


def test_collect(index):
    # Test collecting from an empty index.
    result = index.collect()
    assert isinstance(result, list)
    assert len(result) == 0
    text_obj = TestClass()
    index.add_object(text_obj)
    result = index.collect()
    assert len(result) == 1
    assert result[0] is text_obj


def test_add_object(index):
    text_obj = TestClass(key="value")
    index.add_object(text_obj)
    result = index.collect()
    assert len(result) == 1
    assert result[0] is text_obj


def test_add_object_many(index):
    objs = [TestClass(num=i) for i in range(5)]
    index.add_object_many(objs)
    result = index.collect()
    assert len(result) == 5
    for obj in objs:
        assert obj in result

def test_query(index):
    objs = [TestClass(num=i, active=(i % 2 == 0), score=float(i) * 10.0) for i in range(10)]
    index.add_object_many(objs)

    query = Q.and_(
        Q.eq("active", True),
        Q.gt("score", 50.0)
    )
    result = index.reduced_query(query).collect()
    assert all(obj.active is True and obj.score > 50.0 for obj in result)
    assert len(result) == 2  # Should be objects with num 6 and 8

def test_query_chain(index):
    objs = [TestClass(num=i, active=(i % 2 == 0), score=float(i) * 10.0) for i in range(10)]
    index.add_object_many(objs)

    # First query
    filtered_index = index.reduced(active=True)
    # Second query on the filtered index
    second_filtered_index = filtered_index.reduced_query(Q.gt("score", 50.0))
    res = second_filtered_index.collect()
    assert all(obj.active is True and obj.score > 50.0 for obj in res)
    assert len(res) == 2  # Should be objects with num 6 and 8

    final_filtered_index = second_filtered_index.reduced_query(Q.lt("num", 8))
    final_res = final_filtered_index.collect()
    assert all(obj.active is True and obj.score > 50.0 and obj.num < 8 for obj in final_res)
    assert len(final_res) == 1  # Should be objects with num 6


def test_filtered_index(index):
    objs = [TestClass(num=i, active=(i % 2 == 0), score=float(i) * 10.0) for i in range(10)]
    index.add_object_many(objs)

    # initial filter
    filtered_index = index.reduced(active=True)
    # filter the filtered
    reduced = filtered_index.reduced_query(
        Q.gt("score", 50.0)
    )

    result = reduced.collect()
    assert all(obj.active is True and obj.score > 50.0 for obj in result)
    assert len(result) == 2

def test_and_query(index):
    objs = [TestClass(num=i, active=(i % 2 == 0), score=float(i) * 10.0) for i in range(10)]
    index.add_object_many(objs)

    # Create a nested query
    query = Q.and_(
        Q.eq("active", True),
        Q.or_(
            Q.gt("score", 70.0),
            Q.lt("num", 3)
        )
    )
    result = index.reduced_query(query).collect()
    assert all(obj.active is True and (obj.score > 70.0 or obj.num < 3) for obj in result)
    assert len(result) == 3 # Should be objects with num 0,2,8


def test_nested_object_query(index):
    class NestedTestClass(Indexable):
        pass

    nested_objs = [TestClass(num=i, nested=NestedTestClass(num=i * 10)) for i in range(5)]
    index.add_object_many(nested_objs)

    # Query based on nested object's attribute
    query = Q.eq("nested.num", 20)
    result = index.reduced_query(query).collect()
    assert len(result) == 1
    assert result[0].nested.num == 20

def test_updates_reflect(index):
    obj = TestClass(num=1, active=True)
    index.add_object(obj)

    query = Q.eq("active", True)
    result = index.reduced_query(query).collect()
    assert len(result) == 1

    # Update the object to no longer be active
    obj.active = False

    result = index.reduced_query(query).collect()
    assert len(result) == 0

    # Update back to active
    obj.active = True

    result = index.reduced_query(query).collect()
    assert len(result) == 1



import random
import string
import numpy as np
import pytest

from PyThermite import Index
from PyThermite import Indexable
from PyThermite import QueryExpr as Q


import time
import pandas as pd
import psutil
import os

ITERATIONS = 1

class testclass:
    def __init__(self):
        pass

class Record(Indexable):

    def __repr__(self):
        return (
            f"Record(id={self.id}, age={self.age}, score={self.score:.1f}, "
            f"active={self.active}, country='{self.country}', group='{self.group}', tags='{self.tags}')"
        )
    
    def to_dict(self):
        return {
            "id": self.id,
            "age": self.age,
            "score": self.score,
            "active": self.active,
            "country": self.country,
            "group": self.group,
            "tags": self.tags,
        }


def random_str(length=6):
    return ''.join(random.choices(string.ascii_lowercase, k=length))

# Memory usage helper
def mem_usage_mb():
    return psutil.Process(os.getpid()).memory_info().rss / 1024 / 1024

@pytest.fixture
def prep_data_fixture():
    return prep_data()

def prep_data():
    N = 100_000
    random.seed(42)
    np.random.seed(42)

    print("making build")
    make_data = [
        {
            "id": i,
            "age": np.random.randint(18, 80),
            "score": np.random.rand() * 100,
            "active": np.random.choice([True, False]),
            "country": np.random.choice(["US", "CA", "MX", "FR", "DE"]),
            "group": random_str(),
            "tags": np.random.choice(["a", "b", "c", "d"]),
        }
        for i in range(N)
    ]

    return make_data


def test_performance(prep_data_fixture):
    print("creating Objects")
    start = time.perf_counter()
    data = [
        Record(
            **m
        )
        for m in prep_data_fixture
    ]
    object_build_time = time.perf_counter() - start

    py_objects = [row for row in data]

    start = time.perf_counter()
    cols = ["id", "age", "score", "active", "country", "group", "tags"]
    df = pd.DataFrame(prep_data_fixture, columns=cols)

    pandas_build_time = time.perf_counter() - start
    mem_pix = mem_usage_mb()

    print("Starting Pandas")
    # Benchmark: Pandas
    start = time.perf_counter()
    for i in range(ITERATIONS):
        filtered_df = df[
            (
                (df["age"].between(35, 60)) &
                (df["active"]) &
                (df["score"] > 85.0)
            ) |
            (
                (df["country"].isin(["CA", "MX"])) &
                (df["score"].between(50.0, 75.0)) &
                (df["tags"] != "b")
            )
        ]

        filtered_df = filtered_df[
            (filtered_df["group"] != "guest") &
            (filtered_df["age"] < 65) &
            (filtered_df["country"] != "US")
        ]

    filtered_df.to_dict(orient="records")
    duration_df = time.perf_counter() - start
    mem_df = mem_usage_mb()

    print("Starting building index")
    index = Index()
    start = time.perf_counter()
    index.add_object_many(py_objects)
    duration_bix = time.perf_counter() - start
    mem_bix = mem_usage_mb()

    print("Starting index")
    start = time.perf_counter()
    for i in range(ITERATIONS):
        query = Q.and_(
            Q.or_(
                Q.and_(
                    Q.eq("active", True),
                    Q.bt("age", 35, 60),
                    Q.gt("score", 85.0),
                ),
                Q.and_(
                    Q.in_("country", ["CA", "MX"]),
                    Q.bt("score", 50.0, 75.0),
                    Q.ne("tags", "b"),
                )
            ),
            Q.ne("group", "guest"),
            Q.ne("country", "US"),
            Q.lt("age", 65),
        )
        result = index.reduced_query(query)
    #    filtered_ix = index.reduced(b = 2, a = 1000)
    filtered_ix = result.collect()
    duration_ix = time.perf_counter() - start

    mem_ix = mem_usage_mb()

    # Print Results
    print("\n==== Benchmark Results ====")

    test = index.collect()[0]
    test1 = index.collect()[1]
    test.test_val = test1

    nested_test = index.reduced_query(
        query = Q.eq("test_val.age" ,  test1.age)
    ).collect()

    assert len(nested_test) > 0
    assert len(filtered_df) == len(filtered_ix)

    print(f"Object Build:                   {object_build_time:.4f} s")
    print(f"Pandas Filter Build index:      {pandas_build_time:.4f} s | Mem: {mem_pix:.1f} MB")
    print(f"Pandas Filter:                  {duration_df:.6f} s | Mem: {mem_df:.1f} MB | Result size: {len(filtered_df)}")
    print(f"Your Index Filter Build index:  {duration_bix:.4f} s | Mem: {mem_bix:.1f} MB")
    print(f"Your Index Filter:              {duration_ix:.6f} s | Mem: {mem_ix:.1f} MB | Result size: {len(filtered_ix)}")


# run the pytest with print statements visible
if __name__ == "__main__":
    data = prep_data()
    test_performance(data)



import pytest

from PyThermite import Index, Indexable, QueryExpr as Q, FilteredIndex

class TestClass(Indexable):
    def some_method(self):
        return "Hello from TestClass"


@pytest.fixture
def indexable():
    # Basic creation of an Indexable object.
    return Indexable()


@pytest.fixture
def index():
    return Index()


def test_string(index):
    objs = [TestClass(name=f"object_{i}") for i in range(5)]
    index.add_object_many(objs)

    query = Q.eq("name", "object_3")
    result = index.reduced_query(query)
    assert len(result.collect()) == 1
    assert result.collect()[0].name == "object_3"

    for r in result.collect():
        r.child = TestClass(name="child_of")

    nested_result = index.reduced_query(Q.eq("child.name", "child_of"))
    assert len(nested_result.collect()) == 1
    assert nested_result.collect()[0].name == "object_3"

def test_nest_before_index(index):
    objs = [TestClass(name=f"object_{i}") for i in range(5)]
    for obj in objs:
        obj.child = TestClass(name="child_of", grandchild=TestClass(name="grandchild_of"))
    index.add_object_many(objs)
    nested_result = index.reduced_query(Q.eq("child.name", "child_of"))
    tripple_nested_result = index.reduced_query(Q.eq("child.grandchild.name", "grandchild_of"))

    assert len(nested_result.collect()) == 5
    assert all(r.child.name == "child_of" for r in nested_result.collect())

    assert len(tripple_nested_result.collect()) == 5
    assert all(r.child.grandchild.name == "grandchild_of" for r in tripple_nested_result.collect())


def test_tripple_nest_after_index(index):
    objs = [TestClass(name=f"object_{i}") for i in range(5)]
    index.add_object_many(objs)

    query = Q.eq("name", "object_3")
    result = index.reduced_query(query)
    assert len(result.collect()) == 1
    assert result.collect()[0].name == "object_3"

    for r in result.collect():
        r.child = TestClass(name="child_of")

    nested_result = index.reduced_query(Q.eq("child.name", "child_of"))
    assert len(nested_result.collect()) == 1
    assert nested_result.collect()[0].name == "object_3"

    nested_result.collect()[0].child.grandchild = TestClass(name="grandchild_of")
    tripple_nested_result = index.reduced_query(Q.eq("child.grandchild.name", "grandchild_of"))
    assert len(tripple_nested_result.collect()) == 1
    assert tripple_nested_result.collect()[0].name == "object_3"


def test_filtered_index_chain(index):
    objs = [TestClass(name=f"object_{i}", common=True) for i in range(5)]
    index.add_object_many(objs)

    query = Q.in_("name", ["object_1", "object_3"])
    result = index.reduced_query(query)
    assert len(result.collect()) == 2
    assert all(r.name in ["object_1", "object_3"] for r in result.collect())

    for r in result.collect():
        r.child = TestClass(name="child_of")

    nested_result = index.reduced_query(Q.eq("child.name", "child_of"))
    assert len(nested_result.collect()) == 2
    assert all(r.child.name == "child_of" for r in nested_result.collect())

    tripple_nested_result = nested_result.reduced_query(Q.eq("common", True))
    assert len(tripple_nested_result.collect()) == 2
    assert all(r.child.name == "child_of" for r in nested_result.collect())


def test_nested_object_query_in(index):
    class NestedTestClass(Indexable):
        pass

    nested_objs = [TestClass(num=i, nested=NestedTestClass(num=i * 10)) for i in range(5)]
    index.add_object_many(nested_objs)

    # Query based on nested object's attribute
    query = Q.in_(
        "nested.num", [ 20, 30, 40 ]
    )
    result = index.reduced_query(query).collect()
    assert len(result) == 3
    assert all(obj.nested.num in [20, 30, 40] for obj in result)


def test_nested_object_query_greater(index):
    class NestedTestClass(Indexable):
        pass

    nested_objs = [TestClass(num=i, nested=NestedTestClass(num=i * 10)) for i in range(11)]
    index.add_object_many(nested_objs)

    # Query based on nested object's attribute
    query = Q.gt(
        "nested.num", 50
    )
    result = index.reduced_query(query).collect()
    assert len(result) == 5
    assert all(obj.nested.num in [60, 70, 80, 90, 100] for obj in result)

        # Query based on nested object's attribute
    query = Q.ge(
        "nested.num", 50
    )
    result = index.reduced_query(query).collect()
    assert len(result) == 6
    assert all(obj.nested.num in [50, 60, 70, 80, 90, 100] for obj in result)


def test_nested_object_query_less(index):
    class NestedTestClass(Indexable):
        pass

    nested_objs = [TestClass(num=i, nested=NestedTestClass(num=i * 10)) for i in range(11)]
    index.add_object_many(nested_objs)

    # Query based on nested object's attribute
    query = Q.lt(
        "nested.num", 50
    )
    result = index.reduced_query(query).collect()
    assert len(result) == 5
    assert all(obj.nested.num in [0, 10, 20, 30, 40] for obj in result)

        # Query based on nested object's attribute
    query = Q.le(
        "nested.num", 50
    )
    result = index.reduced_query(query).collect()
    assert len(result) == 6
    assert all(obj.nested.num in [0, 10, 20, 30, 40, 50] for obj in result)


def test_nested_object_query_between(index):
    class NestedTestClass(Indexable):
        pass

    nested_objs = [TestClass(num=i, nested=NestedTestClass(num=i * 10)) for i in range(11)]
    index.add_object_many(nested_objs)

    # Query based on nested object's attribute
    query = Q.bt(
        "nested.num", 50, 90
    )
    result = index.reduced_query(query).collect()
    assert len(result) == 5
    assert all(obj.nested.num in [50, 60, 70, 80, 90] for obj in result)

def test_recursive_ownership(index):
    class Inner(Indexable):
        pass

    class Outer(Indexable):
        pass

    # Each Outer has an Inner, and each Inner also points back to another Outer (recursive chain).
    outers = []
    for i in range(6):
        inner = Inner(num=i * 10)
        outer = Outer(num=i, inner=inner)
        inner.outer = outer
        outers.append(outer)

    # Add objects (this should also add their nested children)
    index.add_object_many(outers)

    # Query across the recursive Inner objects
    query = Q.bt("inner.num", 20, 50)  # should capture 20, 30, 40, 50
    result = index.reduced_query(query).collect()

    pre_inner = result[0].inner
    result[0].inner = 7
    t_res = index.reduced_query(Q.eq("inner.outer.x", "y")).collect()
    assert len(t_res) == 0

    t_res = index.reduced_query(Q.eq("inner", 7)).collect()
    assert len(t_res) == 1
    result[0].inner_other = pre_inner

    result = index.reduced_query(Q.eq("inner.outer.inner.num", 2)).collect()
    assert len(result) == 0 # do not index children

def test_recursive_ownership_1(index):
    class Inner(Indexable):
        pass

    class Outer(Indexable):
        pass

    # Each Outer has an Inner, and each Inner also points back to another Outer (recursive chain).
    outers = []
    for i in range(6):
        inner = Inner(num=i * 10)
        outer = Outer(num=i, inner=inner)
        inner.outer = outer
        outers.append(outer)

    # Add objects (this should also add their nested children)
    index.add_object_many(outers)

    # Query across the recursive Inner objects
    query = Q.bt("inner.num", 20, 50)  # should capture 20, 30, 40, 50
    result = index.reduced_query(query).collect()

    # Ensure we got the correct number of results
    assert len(result) == 4
    nums = [obj.inner.num for obj in result]
    assert all(n in [20, 30, 40, 50] for n in nums)

    # And check that parent Outer.num values align with expectations
    outer_nums = [obj.num for obj in result]
    assert sorted(outer_nums) == [2, 3, 4, 5]

    result = index.reduced_query(Q.eq("inner.num", 20)).collect()
    assert len(result) == 1
    assert result[0].num == 2
    assert result[0].inner.num == 20

    r = index.reduced_query(Q.eq("inner.outer.num", 2)).collect()
    assert len(r) == 0 # do not index children

    result[0].inner.outer.x = "y"
    result = index.reduced_query(Q.eq("x", "y")).collect()
    assert len(result) == 1
    assert result[0].x == "y"
    assert result[0].inner.outer.x == "y"

    result[0].inner = 7
    t_res = index.reduced_query(Q.eq("inner", 7)).collect()
    assert len(t_res) == 1
    assert t_res[0].inner == 7


if __name__ == "__main__":
    test_recursive_ownership_1(Index())
